"""
Warping Invariant Regression using SRVF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf.utility_functions as uf
import fdasrsf.curve_functions as cf
from scipy.interpolate import interp1d
from scipy.optimize import fmin_l_bfgs_b
from scipy.integrate import trapz, cumtrapz
from scipy.linalg import inv, norm, expm
from patsy import bs
from joblib import Parallel, delayed
import ocmlogit_warp as mw
import oclogit_warp as lw

class oc_elastic_regression:
    """
    This class provides elastic regression for for open curves
    using elastic methods
    
    Usage:  obj = oc_elastic_regression(beta,y)

    :param beta: numpy ndarray of shape (n, M, N) describing N curves in R^M
    :param B: optional matrix describing Basis elements
    :param y: numpy array of N responses
    :param alpha: alpha parameter of model
    :param beta: beta(t) of model
    :param betan: aligned curves - numpy ndarray of shape (n,M,N) describing N curves
    in R^M
    :param qn: aligned srvfs - similar structure to betan
    :param gamma: calculated warping functions
    :param q: original training SRSFs
    :param b: basis coefficients
    :param SSE: sum of squared error
    :param nu: regressor curve

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021
    """

    def __init__(self, beta, y, ):
        """
        Construct an instance of the oc_elastic_regression class
        :param beta: numpy ndarray of shape (n, M, N) describing N curves
        :param y: response vector
        :param time: vector of size M describing the sample points
        """

        self.beta = beta
        self.y = y
    
    def calc_model(self, B=None, lam=0, df=40, T=200, max_itr=20, cores=-1):
        """
        This function identifies a regression model for open curves
        using elastic methods

        :param B: optional matrix describing Basis elements
        :param lam: regularization parameter (default 0)
        :param df: number of degrees of freedom B-spline (default 20)
        :param T: number of desired samples along curve (default 100)
        :param max_itr: maximum number of iterations (default 20)
        :param cores: number of cores for parallel processing (default all)
        """
        n = self.beta.shape[0]
        N = self.beta.shape[2]
        time = np.linspace(0, 1, T)

        if n > 500:
            parallel = True
        elif T > 100:
            parallel = True
        else:
            parallel = False

        binsize = np.diff(time)
        binsize = binsize.mean()

        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]

        # second derivative for regularization
        Bdiff = np.zeros((T, Nb))
        for ii in range(0, Nb):
            Bdiff[:, ii] = np.gradient(np.gradient(B[:, ii], binsize), binsize)

        q, beta = preproc_open_curve(self.beta, T)
        self.q = q
        beta0 = beta.copy()
        qn = q.copy()

        gamma = np.tile(np.linspace(0, 1, T), (N, 1))
        gamma = gamma.transpose()
        O_hat = np.tile(np.eye(n), (N, 1, 1)).T

        itr = 1
        self.SSE = np.zeros(max_itr)
        while itr <= max_itr:
            print("Iteration: %d" % itr)
            # align data

            # OLS using basis
            Phi = np.ones((N, n * Nb + 1))
            for ii in range(0, N):
                for jj in range(0, n):
                    for kk in range(1, Nb + 1):
                        Phi[ii, jj * Nb + kk] = trapz(qn[jj, :, ii] * B[:, kk - 1], time)
            
            R = np.zeros((n * Nb+1, n * Nb+1))
            for kk in range(0, n):
                for ii in range(1, Nb+1):
                    for jj in range(1, Nb+1):
                        R[kk * Nb + ii, kk * Nb + jj] = trapz(Bdiff[:, ii-1] * Bdiff[:, jj-1], time)

            xx = np.dot(Phi.T, Phi)
            inv_xx = inv(xx + lam * R)
            xy = np.dot(Phi.T, self.y)
            b = np.dot(inv_xx, xy)

            alpha = b[0]
            nu = np.zeros((n, T))
            for ii in range(0, n):
                nu[ii, :] = B.dot(b[(ii * Nb + 1):((ii + 1) * Nb + 1)])

            # compute the SSE
            int_X = np.zeros(N)
            for ii in range(0, N):
                int_X[ii] = cf.innerprod_q2(qn[:, :, ii], nu)

            self.SSE[itr - 1] = sum((self.y.reshape(N) - alpha - int_X) ** 2)

            # find gamma
            gamma_new = np.zeros((T, N))
            if parallel:
                out = Parallel(n_jobs=cores)(delayed(regression_warp)(nu, q[:, :, n], self.y[n], alpha) for n in range(N))
                for ii in range(0, N):
                    gamma_new[:, ii] = out[ii][0]
                    beta1n = cf.group_action_by_gamma_coord(out[ii][1].dot(beta0[:, :, ii]), out[ii][0])
                    beta[:, :, ii] = beta1n
                    O_hat[:, :, ii] = out[ii][1]
                    qn[:, :, ii] = cf.curve_to_q(beta1n)[0]
            else:
                for ii in range(0, N):
                    q1 = q[:, :, ii]
                    gammatmp, Otmp = regression_warp(nu, q1, self.y[ii], alpha)
                    gamma_new[:, ii] = gammatmp
                    beta1n = cf.group_action_by_gamma_coord(Otmp.dot(beta0[:, :, ii]), gammatmp)
                    beta[:, :, ii] = beta1n
                    O_hat[:, :, ii] = Otmp
                    qn[:, :, ii] = cf.curve_to_q(beta1n)[0]


            if np.abs(self.SSE[itr - 1] - self.SSE[itr - 2]) < 1e-15:
                break
            else:
                gamma = gamma_new

            itr += 1

        tau = np.zeros(N)
        self.alpha = alpha
        self.nu = nu
        self.beta0 = beta0
        self.betan = beta
        self.gamma = gamma
        self.qn = qn
        self.B = B
        self.O = O_hat
        self.b = b[1:-1]
        self.SSE = self.SSE[0:itr]

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param beta: numpy ndarray of shape (M,N) of M functions with N samples
        :param y: truth if available
        """ 

        if newdata != None:
            beta = newdata['beta']
            y = newdata['y']

            T = self.q.shape[1]
            n = beta.shape[2]
            N = self.q.shape[2]

            q, beta = preproc_open_curve(beta, T)

            y_pred = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - q[:, :, ii][:, :, np.newaxis]
                # dist = np.linalg.norm(np.abs(diff), axis=(0, 1)) ** 2
                dist = np.zeros(N)
                for jj in range(0, N):
                    dist[jj] = np.linalg.norm(np.abs(diff[:, :, jj])) ** 2
                beta1 = beta[:, :, ii]
                beta1 = self.O[:, :, dist.argmin()].dot(beta1)
                beta1 = cf.group_action_by_gamma_coord(beta1,
                                                       self.gamma[:, dist.argmin()])
                q_tmp = cf.curve_to_q(beta1)[0]

                y_pred[ii] = self.alpha + cf.innerprod_q2(q_tmp, self.nu)
            
            if y is None:
                self.SSE = None
            else:
                self.SSE = sum((y - y_pred) ** 2)
            
            self.y_pred = y_pred

        else:
            n = self.q.shape[2]
            N = self.q.shape[2]
            y_pred = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - self.q[:, :, ii][:, :, np.newaxis]
                # dist = np.linalg.norm(np.abs(diff), axis=(0, 1)) ** 2
                dist = np.zeros(N)
                for jj in range(0, N):
                    dist[jj] = np.linalg.norm(np.abs(diff[:, :, jj])) ** 2
                beta1 = self.beta0[:, :, ii]
                beta1 = self.O[:, :, dist.argmin()].dot(beta1)
                beta1 = cf.group_action_by_gamma_coord(beta1,
                                                       self.gamma[:, dist.argmin()])
                q_tmp = cf.curve_to_q(beta1)[0]

                y_pred[ii] = self.alpha + cf.innerprod_q2(q_tmp, self.nu)
            
            self.y_pred = y_pred

        return


class oc_elastic_logistic:
    """
    This class identifies a logistic regression model with
    phase-variability using elastic methods for open curves

    Usage:  obj = oc_elastic_logistic(beta,y)

    :param beta: numpy ndarray of shape (n, M, N) describing N curves
    in R^M
    :param y: numpy array of N responses
    :param B: optional matrix describing Basis elements
    :param alpha: alpha parameter of model
    :param nu: nu(t) of model
    :param betan: aligned curves - numpy ndarray of shape (n,T,N)
    :param O: calculated rotation matrices
    :param gamma: calculated warping functions
    :param B: basis matrix
    :param b: basis coefficients
    :param Loss: logistic loss

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021

    """

    def __init__(self, beta, y, ):
        """
        Construct an instance of the elastic_regression class
        :param beta: numpy ndarray of shape (n, M, N) describing N curves
        :param y: response vector
        :param time: vector of size M describing the sample points
        """

        self.beta = beta
        self.y = y
    
    def calc_model(self, B=None, df=60, T=100, max_itr=40, cores=-1,
                    deltaO=.1, deltag=.05, method=1):
        """
        This function identifies a logistic regression model with
        phase-variability using elastic methods for open curves

        :param B: optional matrix describing Basis elements
        :param df: number of degrees of freedom B-spline (default 20)
        :param T: number of desired samples along curve (default 100)
        :param max_itr: maximum number of iterations (default 20)
        :param cores: number of cores for parallel processing (default all)
        
        """
        n = self.beta.shape[0]
        N = self.beta.shape[2]
        time = np.linspace(0, 1, T)

        if n > 500:
            parallel = True
        elif T > 100:
            parallel = True
        else:
            parallel = True

        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]

        q, beta = preproc_open_curve(self.beta, T)
        beta0 = beta.copy()
        qn = q.copy()

        gamma = np.tile(np.linspace(0, 1, T), (N, 1))
        gamma = gamma.transpose()
        O_hat = np.tile(np.eye(n), (N, 1, 1)).T

        itr = 1
        LL = np.zeros(max_itr + 1)
        while itr <= max_itr:
            print("Iteration: %d" % itr)

            Phi = np.ones((N, n * Nb + 1))
            for ii in range(0, N):
                for jj in range(0, n):
                    for kk in range(1, Nb + 1):
                        Phi[ii, jj * Nb + kk] = trapz(qn[jj, :, ii] * B[:, kk - 1], time)

            # Find alpha and beta using l_bfgs
            b0 = np.zeros(n * Nb + 1)
            out = fmin_l_bfgs_b(logit_loss, b0, fprime=logit_gradient,
                                args=(Phi, self.y), pgtol=1e-10, maxiter=200,
                                maxfun=250, factr=1e-30)
            b = out[0]
            b = b/norm(b)
            # alpha_norm = b1[0]
            alpha = b[0]
            nu = np.zeros((n, T))
            for ii in range(0, n):
                nu[ii, :] = B.dot(b[(ii * Nb + 1):((ii + 1) * Nb + 1)])

            # compute the logistic loss
            LL[itr] = logit_loss(b, Phi, self.y)

            # find gamma
            gamma_new = np.zeros((T, N))
            if parallel:
                out = Parallel(n_jobs=cores)(delayed(logistic_warp)(alpha, nu, q[:, :, ii], self.y[ii], deltaO=deltaO, deltag=deltag, method=method) for ii in range(N))
                for ii in range(0, N):
                    gamma_new[:, ii] = out[ii][0]
                    beta1n = cf.group_action_by_gamma_coord(out[ii][1].dot(beta0[:, :, ii]), out[ii][0])
                    beta[:, :, ii] = beta1n
                    O_hat[:, :, ii] = out[ii][1]
                    qn[:, :, ii] = cf.curve_to_q(beta[:, :, ii])[0]
            else:
                for ii in range(0, N):
                    q1 = q[:, :, ii]
                    gammatmp, Otmp, tautmp = logistic_warp(alpha, nu, q1, self.y[ii], deltaO=deltaO, 
                                                           deltag=deltag, method=method)
                    gamma_new[:, ii] = gammatmp
                    beta1n = cf.group_action_by_gamma_coord(Otmp.dot(beta0[:, :, ii]), gammatmp)
                    beta[:, :, ii] = beta1n
                    O_hat[:, :, ii] = Otmp
                    qn[:, :, ii] = cf.curve_to_q(beta[:, :, ii])[0]

            if norm(gamma - gamma_new) < 1e-5:
                break
            else:
                gamma = gamma_new.copy()

            itr += 1

        tau = np.zeros(N)
        self.alpha = alpha
        self.nu = nu
        self.betan = beta
        self.beta0 = beta0
        self.gamma = gamma_new
        self.q = q
        self.O = O_hat
        self.tau = tau
        self.B = B
        self.b = b[1:-1]
        self.Loss = LL[1:itr]
        self.qn = qn

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param beta: numpy ndarray of shape (M,N) of M functions with N samples
        :param y: truth if available
        """ 

        if newdata != None:
            beta = newdata['beta']
            y = newdata['y']

            T = self.q.shape[1]
            n = beta.shape[2]
            N = self.q.shape[2]

            q, beta = preproc_open_curve(beta, T)

            y_pred = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - q[:, :, ii][:, :, np.newaxis]
                dist = np.zeros(N)
                for jj in range(0, N):
                    dist[jj] = np.linalg.norm(np.abs(diff[:, :, jj])) ** 2
                beta1 = beta[:, :, ii]
                beta1 = self.O[:, :, dist.argmin()].dot(beta1)
                beta1 = cf.group_action_by_gamma_coord(beta1,
                                                       self.gamma[:, dist.argmin()])
                q_tmp = cf.curve_to_q(beta1)[0]

                y_pred[ii] = self.alpha + cf.innerprod_q2(q_tmp, self.nu)
            
            if y is None:
                y_pred = phi(y_pred)
                y_labels = np.ones(n)
                y_labels[y_pred < 0.5] = -1
                self.PC = None
            else:
                y_pred = phi(y_pred)
                y_labels = np.ones(n)
                y_labels[y_pred < 0.5] = -1
                TP = sum(y[y_labels == 1] == 1)
                FP = sum(y[y_labels == -1] == 1)
                TN = sum(y[y_labels == -1] == -1)
                FN = sum(y[y_labels == 1] == -1)
                self.PC = (TP + TN) / float(TP + FP + FN + TN)
            
            self.y_pred = y_pred
            self.y_labels = y_labels

        else:
            n = self.q.shape[2]
            N = self.q.shape[2]
            y_pred = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - self.q[:, :, ii][:, :, np.newaxis]
                dist = np.zeros(N)
                for jj in range(0, N):
                    dist[jj] = np.linalg.norm(np.abs(diff[:, :, jj])) ** 2
                beta1 = self.beta0[:, :, ii]
                beta1 = self.O[:, :, dist.argmin()].dot(beta1)
                beta1 = cf.group_action_by_gamma_coord(beta1,
                                                       self.gamma[:, dist.argmin()])
                q_tmp = cf.curve_to_q(beta1)[0]

                y_pred[ii] = self.alpha + cf.innerprod_q2(q_tmp, self.nu)

            y_pred = phi(y_pred)
            y_labels = np.ones(n)
            y_labels[y_pred < 0.5] = -1
            TP = sum(self.y[y_labels == 1] == 1)
            FP = sum(self.y[y_labels == -1] == 1)
            TN = sum(self.y[y_labels == -1] == -1)
            FN = sum(self.y[y_labels == 1] == -1)
            self.PC = (TP + TN) / float(TP + FP + FN + TN)
            
            self.y_pred = y_pred
            self.y_labels = y_labels

        return



class oc_elastic_mlogistic:
    """
    This class identifies a multinomial logistic regression model with
    phase-variability using elastic methods for open curves

    Usage:  obj = oc_elastic_logistic(beta,y)

    :param beta: numpy ndarray of shape (n, M, N) describing N curves
    in R^M
    :param y: numpy array of N responses
    :param B: optional matrix describing Basis elements
    :param alpha: alpha parameter of model
    :param nu: nu(t) of model
    :param betan: aligned curves - numpy ndarray of shape (n,T,N)
    :param O: calculated rotation matrices
    :param gamma: calculated warping functions
    :param B: basis matrix
    :param b: basis coefficients
    :param Loss: logistic loss

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021

    """
    def __init__(self, beta, y, ):
        """
        Construct an instance of the elastic_regression class
        :param beta: numpy ndarray of shape (n, M, N) describing N curves
        :param y: response vector
        :param time: vector of size M describing the sample points
        """

        self.beta = beta
        self.y = y

        # Code labels
        m = y.max()
        N = beta.shape[2]
        Y = np.zeros((N, m), dtype=int)
        for ii in range(0, N):
            Y[ii, y[ii]-1] = 1
        self.Y = Y
    
    
    def calc_model(self, B=None, df=20, T=100, max_itr=30, cores=-1,
                   deltaO=.003, deltag=.003):
        """
        This function identifies a multinomial logistic regression model with
        phase-variability using elastic methods for open curves

        :param B: optional matrix describing Basis elements
        :param df: number of degrees of freedom B-spline (default 20)
        :param T: number of desired samples along curve (default 100)
        :param max_itr: maximum number of iterations (default 20)
        :param cores: number of cores for parallel processing (default all)

        """
        n = self.beta.shape[0]
        N = self.beta.shape[2]
        time = np.linspace(0, 1, T)
        m = self.y.max()

        if n > 500:
            parallel = True
        elif T > 100:
            parallel = True
        else:
            parallel = True


        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]

        q, beta = preproc_open_curve(self.beta, T)
        qn = q.copy()
        beta0 = beta.copy()

        gamma = np.tile(np.linspace(0, 1, T), (N, 1))
        gamma = gamma.transpose()
        O_hat = np.tile(np.eye(n), (N, 1, 1)).T

        itr = 1
        LL = np.zeros(max_itr+1)
        while itr <= max_itr:
            print("Iteration: %d" % itr)

            Phi = np.ones((N, n * Nb + 1))
            for ii in range(0, N):
                for jj in range(0, n):
                    for kk in range(1, Nb + 1):
                        Phi[ii, jj * Nb + kk] = trapz(qn[jj, :, ii] * B[:, kk - 1], time)

            # Find alpha and beta using l_bfgs
            b0 = np.zeros(m * (n * Nb + 1))
            out = fmin_l_bfgs_b(mlogit_loss, b0, fprime=mlogit_gradient,
                                args=(Phi, self.Y), pgtol=1e-10, maxiter=200,
                                maxfun=250, factr=1e-30)
            b = out[0]
            B0 = b.reshape(n * Nb + 1, m)
            alpha = B0[0, :]
            nu = np.zeros((n, T, m))
            for i in range(0, m):
                for j in range(0, n):
                    nu[j, :, i] = B.dot(B0[(j * Nb + 1):((j + 1) * Nb + 1), i])

            # compute the logistic loss
            LL[itr] = mlogit_loss(b, Phi, self.Y)

            # find gamma
            gamma_new = np.zeros((T, N))
            if parallel:
                out = Parallel(n_jobs=cores)(delayed(mlogit_warp_grad)(alpha, nu, q[:, :, n], self.Y[n, :], deltaO=deltaO, deltag=deltag) for n in range(N))
                for ii in range(0, N):
                    gamma_new[:, ii] = out[ii][0]
                    beta1n = cf.group_action_by_gamma_coord(out[ii][1].dot(beta0[:, :, ii]), out[ii][0])
                    beta[:, :, ii] = beta1n
                    O_hat[:, :, ii] = out[ii][1]
                    qn[:, :, ii] = cf.curve_to_q(beta[:, :, ii])[0]
            else:
                for ii in range(0, N):
                    gammatmp, Otmp = mlogit_warp_grad(alpha, nu, q[:, :, ii], self.Y[ii, :], deltaO=deltaO, deltag=deltag)
                    gamma_new[:, ii] = gammatmp
                    beta1n = cf.group_action_by_gamma_coord(Otmp.dot(beta0[:, :, ii]), gammatmp)
                    beta[:, :, ii] = beta1n
                    O_hat[:, :, ii] = Otmp
                    qn[:, :, ii] = cf.curve_to_q(beta[:, :, ii])[0]

            if norm(gamma - gamma_new) < 1e-5:
                break
            else:
                gamma = gamma_new.copy()

            itr += 1
        
        self.alpha = alpha
        self.nu = nu
        self.beta0 = beta0
        self.betan = beta
        self.q = q
        self.qn = qn
        self.gamma = gamma_new
        self.O = O_hat
        self.B = B
        self.b = b[1:-1]
        self.Loss = LL[1:itr]
        self.n_classes = m

        return 


    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param beta: numpy ndarray of shape (M,N) of M functions with N samples
        :param y: truth if available
        """ 

        if newdata != None:
            beta = newdata['beta']
            y = newdata['y']

            T = self.q.shape[1]
            n = beta.shape[2]
            N = self.q.shape[2]

            q, beta = preproc_open_curve(beta, T)

            m = self.n_classes
            y_pred = np.zeros((n, m))
            for ii in range(0, n):
                diff = self.q - q[:, :, ii][:, :, np.newaxis]
                dist = np.zeros(N)
                for jj in range(0, N):
                    dist[jj] = np.linalg.norm(np.abs(diff[:, :, jj])) ** 2
                beta1 = beta[:, :, ii]
                beta1 = self.O[:, :, dist.argmin()].dot(beta1)
                beta1 = cf.group_action_by_gamma_coord(beta1,
                                                       self.gamma[:, dist.argmin()])
                q_tmp = cf.curve_to_q(beta1)[0]

                for jj in range(0, m):
                    y_pred[ii, jj] = self.alpha[jj] + cf.innerprod_q2(q_tmp, self.nu[:, :, jj])
            
            if y is None:
                y_pred = phi(y_pred.ravel())
                y_pred = y_pred.reshape(n, m)
                y_labels = y_pred.argmax(axis=1) + 1
                self.PC = None
            else:
                y_pred = phi(y_pred.ravel())
                y_pred = y_pred.reshape(n, m)
                y_labels = y_pred.argmax(axis=1) + 1
                PC = np.zeros(m)
                cls_set = np.arange(1, m + 1)
                for ii in range(0, m):
                    cls_sub = np.delete(cls_set, ii)
                    TP = sum(y[y_labels == (ii + 1)] == (ii + 1))
                    FP = sum(y[np.in1d(y_labels, cls_sub)] == (ii + 1))
                    TN = sum(y[np.in1d(y_labels, cls_sub)] ==
                            y_labels[np.in1d(y_labels, cls_sub)])
                    FN = sum(np.in1d(y[y_labels == (ii + 1)], cls_sub))
                    PC[ii] = (TP + TN) / float(TP + FP + FN + TN)

                self.PC = sum(y == y_labels) / float(y_labels.size)
            
            self.y_pred = y_pred
            self.y_labels = y_labels

        else:
            n = self.q.shape[2]
            N = self.q.shape[2]
            m = self.n_classes
            y_pred = np.zeros((n, m))
            for ii in range(0, n):
                diff = self.q - self.q[:, :, ii][:, :, np.newaxis]
                dist = np.zeros(N)
                for jj in range(0, N):
                    dist[jj] = np.linalg.norm(np.abs(diff[:, :, jj])) ** 2
                beta1 = self.beta0[:, :, ii]
                beta1 = self.O[:, :, dist.argmin()].dot(beta1)
                beta1 = cf.group_action_by_gamma_coord(beta1,
                                                       self.gamma[:, dist.argmin()])
                q_tmp = cf.curve_to_q(beta1)[0]

                for jj in range(0, m):
                    y_pred[ii, jj] = self.alpha[jj] + cf.innerprod_q2(q_tmp, self.nu[:, :, jj])

            y_pred = phi(y_pred.ravel())
            y_pred = y_pred.reshape(n, m)
            y_labels = y_pred.argmax(axis=1) + 1
            PC = np.zeros(m)
            cls_set = np.arange(1, m + 1)
            for ii in range(0, m):
                cls_sub = np.delete(cls_set, ii)
                TP = sum(y[y_labels == (ii + 1)] == (ii + 1))
                FP = sum(y[np.in1d(y_labels, cls_sub)] == (ii + 1))
                TN = sum(y[np.in1d(y_labels, cls_sub)] ==
                        y_labels[np.in1d(y_labels, cls_sub)])
                FN = sum(np.in1d(y[y_labels == (ii + 1)], cls_sub))
                PC[ii] = (TP + TN) / float(TP + FP + FN + TN)

            self.PC = sum(self.y == y_labels) / float(y_labels.size)
            
            self.y_pred = y_pred
            self.y_labels = y_labels

        return


# helper function for curve manipulation
def preproc_open_curve(beta, T=100):
    n, M, k = beta.shape

    q = np.zeros((n, T, k))
    beta2 = np.zeros((n, T, k))
    for i in range(0, k):
        beta1 = beta[:, :, i]
        beta1, scale = cf.scale_curve(beta1)
        if T != M:
            beta1 = cf.resamplecurve(beta1, T)
        centroid1 = cf.calculatecentroid(beta1)
        beta1 = beta1 - np.tile(centroid1, [T, 1]).T
        beta2[:, :, i] = beta1
        q[:, :, i] = cf.curve_to_q(beta1)[0]

    return (q, beta2)


# helper functions for linear regression
def regression_warp(nu, q, y, alpha):
    """
    calculates optimal warping for function linear regression

    :param nu: numpy ndarray of srvf (M,N) of M functions with N samples
    :param q: numpy ndarray of srvf (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses
    :param alpha: numpy scalar

    :rtype: numpy array
    :return gamma_new: warping function

    """
    T = q.shape[1]

    qM, O_M, gam_M = cf.find_rotation_and_seed_q(nu, q, rotation=False)
    y_M = cf.innerprod_q2(qM, nu)

    qm, O_m, gam_m = cf.find_rotation_and_seed_q(-1 * nu, q, rotation=False)
    y_m = cf.innerprod_q2(qm, nu)

    if y > alpha + y_M:
        O_hat = O_M
        gamma_new = gam_M
    elif y < alpha + y_m:
        O_hat = O_m
        gamma_new = gam_m
    else:
        gamma_new, O_hat = cf.curve_zero_crossing(y - alpha, q, nu, y_M, y_m, gam_M,
                                                  gam_m)

    return(gamma_new, O_hat)


# helper functions for logistic regression
def logistic_warp(alpha, nu, q, y, deltaO=.1, deltag=.05, max_itr=8000,
                  tol=1e-4, display=0, method=1):
    """
    calculates optimal warping for function logistic regression

    :param alpha: scalar
    :param nu: numpy ndarray of shape (M,N) of M functions with N samples
    :param q: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses

    :rtype: numpy array
    :return gamma: warping function

    """
    if method == 1:
        tau = 0
        # q, scale = cf.scale_curve(q)
        q = q/norm(q)
        # nu, scale = cf.scale_curve(nu)
        # alpha = alpha/scale

        gam_old, O_old = lw.oclogit_warp(np.ascontiguousarray(alpha),
                                         np.ascontiguousarray(nu),
                                         np.ascontiguousarray(q),
                                         np.ascontiguousarray(y, dtype=np.int32),
                                         max_itr, tol, deltaO, deltag, display)
    elif method == 2:
        betanu = cf.q_to_curve(nu)
        beta = cf.q_to_curve(q)
        T = beta.shape[1]
        if y == 1:
            beta1, O_old, tau = cf.find_rotation_and_seed_coord(betanu, beta)
            q = cf.curve_to_q(beta1)[0]
            gam_old = cf.optimum_reparam_curve(nu, q)
        elif y == -1:
            beta1, O_old, tau = cf.find_rotation_and_seed_coord(-1 * betanu, beta)
            q = cf.curve_to_q(beta1)[0]
            gam_old = cf.optimum_reparam_curve(-1 * nu, q)


    return (gam_old, O_old, tau)


def phi(t):
    """
    calculates logistic function, returns 1 / (1 + exp(-t))

    :param t: scalar

    :rtype: numpy array
    :return out: return value

    """
    # logistic function, returns 1 / (1 + exp(-t))
    idx = t > 0
    out = np.empty(t.size, dtype=np.float)
    out[idx] = 1. / (1 + np.exp(-t[idx]))
    exp_t = np.exp(t[~idx])
    out[~idx] = exp_t / (1. + exp_t)
    return out


def logit_loss(b, X, y, lam=0.0):
    """
    logistic loss function, returns Sum{-log(phi(t))}

    :param b: numpy ndarray of shape (M,N) of M functions with N samples
    :param X: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses

    :rtype: numpy array
    :return out: loss value

    """
    z = X.dot(b)
    yz = y * z
    idx = yz > 0
    out = np.zeros_like(yz)
    out[idx] = np.log(1 + np.exp(-yz[idx]))
    out[~idx] = (-yz[~idx] + np.log(1 + np.exp(yz[~idx])))
    out = out.sum() + .5 * lam * b.dot(b)
    return out


def logit_gradient(b, X, y, lam=0.0):
    """
    calculates gradient of the logistic loss

    :param b: numpy ndarray of shape (M,N) of M functions with N samples
    :param X: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses

    :rtype: numpy array
    :return grad: gradient of logistic loss

    """
    z = X.dot(b)
    z = phi(y * z)
    z0 = (z - 1) * y
    grad = X.T.dot(z0) + lam * b
    return grad


def logit_hessian(s, b, X, y):
    """
    calculates hessian of the logistic loss

    :param s: numpy ndarray of shape (M,N) of M functions with N samples
    :param b: numpy ndarray of shape (M,N) of M functions with N samples
    :param X: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses

    :rtype: numpy array
    :return out: hessian of logistic loss

    """
    z = X.dot(b)
    z = phi(y * z)
    d = z * (1 - z)
    wa = d * X.dot(s)
    Hs = X.T.dot(wa)
    out = Hs
    return out


# helper functions for multinomial logistic regression
def mlogit_warp_grad(alpha, nu, q, y, max_itr=8000, tol=1e-4,
                     deltaO=0.008, deltag=0.008, display=0):
    """
    calculates optimal warping for functional multinomial logistic regression

    :param alpha: scalar
    :param nu: numpy ndarray of shape (M,N) of M functions with N samples
    :param q: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses
    :param max_itr: maximum number of iterations (Default=8000)
    :param tol: stopping tolerance (Default=1e-10)
    :param deltaO: gradient step size for rotation (Default=0.008)
    :param deltag: gradient step size for warping (Default=0.008)
    :param display: display iterations (Default=0)

    :rtype: tuple of numpy array
    :return gam_old: warping function

    """

    alpha = alpha/norm(alpha)
    q, scale = cf.scale_curve(q)  # q/norm(q)
    for ii in range(0, nu.shape[2]):
        nu[:, :, ii], scale = cf.scale_curve(nu[:, :, ii])  # nu/norm(nu)

    gam_old, O_old = mw.ocmlogit_warp(np.ascontiguousarray(alpha),
                                      np.ascontiguousarray(nu),
                                      np.ascontiguousarray(q),
                                      np.ascontiguousarray(y, dtype=np.int32),
                                      max_itr, tol, deltaO, deltag, display)

    return (gam_old, O_old)


def mlogit_loss(b, X, Y):
    """
    calculates multinomial logistic loss (negative log-likelihood)

    :param b: numpy ndarray of shape (M,N) of M functions with N samples
    :param X: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses

    :rtype: numpy array
    :return nll: negative log-likelihood

    """
    N, m = Y.shape  # n_samples, n_classes
    M = X.shape[1]  # n_features
    B = b.reshape(M, m)
    Yhat = np.dot(X, B)
    Yhat -= Yhat.min(axis=1)[:, np.newaxis]
    Yhat = np.exp(-Yhat)
    # l1-normalize
    Yhat /= Yhat.sum(axis=1)[:, np.newaxis]

    Yhat = Yhat * Y
    nll = np.sum(np.log(Yhat.sum(axis=1)))
    nll /= -float(N)

    return nll


def mlogit_gradient(b, X, Y):
    """
    calculates gradient of the multinomial logistic loss

    :param b: numpy ndarray of shape (M,N) of M functions with N samples
    :param X: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses

    :rtype: numpy array
    :return grad: gradient

    """
    N, m = Y.shape  # n_samples, n_classes
    M = X.shape[1]  # n_features
    B = b.reshape(M, m)
    Yhat = np.dot(X, B)
    Yhat -= Yhat.min(axis=1)[:, np.newaxis]
    Yhat = np.exp(-Yhat)
    # l1-normalize
    Yhat /= Yhat.sum(axis=1)[:, np.newaxis]

    _Yhat = Yhat * Y
    _Yhat /= _Yhat.sum(axis=1)[:, np.newaxis]
    Yhat -= _Yhat
    grad = np.dot(X.T, Yhat)
    grad /= -float(N)
    grad = grad.ravel()

    return grad



"""
Warping Invariant GML Regression using SRSF

moduleauthor:: Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
from patsy import bs
from scipy.optimize import minimize
from numpy.random import rand
from joblib import Parallel, delayed

class elastic_glm_regression:
    """
    This class provides elastic glm regression for functional data using the
    SRVF framework accounting for warping
    
    Usage:  obj = elastic_glm_regression(f,y,time)
    
    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N
    :param time: time vector of length M
    :param alpha: intercept
    :param b: coefficient vector
    :param B: basis matrix
    :param lambda: regularization parameter
    :param SSE: sum of squared errors

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_glm_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception('Columns of f and time must be equal')

        self.f = f
        self.y = y
        self.time = time

    def calc_model(self, link='linear', B=None, lam=0, df=20, max_itr=20, smooth_data=False, sparam=25, parallel=False):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param link: string of link function ('linear', 'quadratic', 'cubic')
        :param B: optional matrix describing Basis elements
        :param lam: regularization parameter (default 0)
        :param df: number of degrees of freedom B-spline (default 20)
        :param max_itr: maximum number of iterations (default 20)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: run in parallel (default = F)
        """
        if smooth_data:
            self.f = fs.smooth_data(self.f,sparam)
        
        print("Link: %s" % link)
        print("Lambda: %5.1f" % lam)

        self.lam = lam
        self.link = link

        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(self.time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]
        self.B = B

        n = self.f.shape[1]

        print("Initializing")
        b0 = rand(Nb+1)        
        out = minimize(MyLogLikelihoodFn, b0, args=(self.y,self.B,self.time,self.f,parallel), method="SLSQP")

        a = out.x

        if self.link == 'linear':
            h1, c_hat, cost = Amplitude_Index(self.f, self.time, self.B, self.y, max_itr, a, 1, parallel)
            yhat1 = c_hat[0] + MapC_to_y(n,c_hat[1:],self.B,self.time,self.f,parallel)
            yhat = np.polyval(h1,yhat1)
        elif self.link == 'quadratic':
            h1, c_hat, cost = Amplitude_Index(self.f, self.time, self.B, self.y, max_itr, a, 2, parallel)
            yhat1 = c_hat[0] + MapC_to_y(n,c_hat[1:],self.B,self.time,self.f,parallel)
            yhat = np.polyval(h1,yhat1)
        elif self.link == 'cubic':
            h1, c_hat, cost = Amplitude_Index(self.f, self.time, self.B, self.y, max_itr, a, 3, parallel)
            yhat1 = c_hat[0] + MapC_to_y(n,c_hat[1:],self.B,self.time,self.f,parallel)
            yhat = np.polyval(h1,yhat1)
        else:
            raise Exception('Invalid Link')
        
        tmp = (self.y-yhat)**2
        self.SSE = tmp.sum()
        self.h = h1
        self.alpha = c_hat[0]
        self.b = c_hat[1:]

        return

    def predict(self, newdata=None, parallel=True):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        if newdata != None:
            f = newdata['f']
            time = newdata['time']
            y = newdata['y']
            sparam = newdata['sparam']
            if newdata['smooth']:
                f = fs.smooth_data(f,sparam)

            n = f.shape[1]
            yhat1 = self.alpha + MapC_to_y(n,self.b,self.B,time,f,parallel)
            yhat = np.polyval(self.h,yhat1)

            if y is None:
                self.SSE = np.nan
            else:
                self.SSE = np.sum((y-yhat)**2)
            
            self.y_pred = yhat

        else:
            n = self.f.shape[1]
            yhat1 = self.alpha + MapC_to_y(n,self.b,self.B,self.time,self.f,parallel)
            yhat = np.polyval(self.h,yhat1)
            self.SSE = np.sum((self.y-yhat)**2)
            self.y_pred = yhat

        return


def Amplitude_Index(f, t, B, y0, MaxIter, b, link, parallel):

    J = B.shape[1]
    n = f.shape[1]
    c_hat = b
    cost = 10000
    itr = 0
    while itr < MaxIter:
        itr += 1
        print("updating step: iter=%d" % (itr))
        y = c_hat[0] + MapC_to_y(n,c_hat[1:],B,t,f,parallel)
        h = np.polyfit(y, y0, link)
        b0 = rand(J+1)
        out = minimize(MyLogLikelihoodFn2, b0, args=(y0,B,t,f,h,parallel), method="SLSQP")

        if cost > out.fun:
            cost = out.fun
            c_hat = out.x
        else:
            c_hat = out.x
            break

    return(h,c_hat,cost)

def MyLogLikelihoodFn2(c, y0, B, t, f, h, parallel):

    N = f.shape[1]
    J = c.shape[0]
    y = c[0] + MapC_to_y(N,c[1:],B,t,f,parallel)
    tmp = np.polyval(h,y)
    x = (y0-tmp)*(y0-tmp)

    return(x.sum())

def MyLogLikelihoodFn(c, y0, B, t, f, parallel):

    N = f.shape[1]
    J = c.shape[0]
    y = c[0] + MapC_to_y(N,c[1:],B,t,f,parallel)
    x = (y0-y)*(y0-y)

    return(x.sum())
        
def MapC_to_y(n,c,B,t,f,parallel):

    dt = np.diff(t)
    dt = dt.mean()

    y = np.zeros(n)

    if parallel:
        bet = np.dot(B,c)
        q1 = uf.f_to_srsf(bet, t)
        y = Parallel(n_jobs=-1)(delayed(map_driver)(q1,
                                            f[:,k], bet, t, dt) for k in range(n))
    else:
        for k in range(0,n):
            bet = np.dot(B,c)
            q1 = uf.f_to_srsf(bet, t)
            q2 = uf.f_to_srsf(f[:,k], t)
            gam = uf.optimum_reparam(q1,t,q2)
            fn = uf.warp_f_gamma(t, f[:,k], gam)
            tmp = bet*fn
            y[k] = tmp.sum()*dt
    
    return(y)

def map_driver(q1, f, bet, t, dt):
    q2 = uf.f_to_srsf(f, t)
    gam = uf.optimum_reparam(q1,t,q2)
    fn = uf.warp_f_gamma(t, f, gam)
    tmp = bet*fn
    y = tmp.sum()*dt

    return y



"""
geometry functions for SRSF Manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

from numpy import arccos, sin, cos, linspace, zeros, sqrt, finfo, double
from numpy import ones, diff, gradient
from scipy.integrate import trapz, cumtrapz


def inv_exp_map(Psi, psi):
    tmp = inner_product(Psi,psi)
    if tmp > 1:
        tmp = 1
    if tmp < -1:
        tmp = -1

    theta = arccos(tmp)

    if (theta < 1e-10):
        exp_inv = zeros(psi.shape[0])
    else: 
        exp_inv = theta / sin(theta) * (psi - cos(theta)*Psi)
    
    return exp_inv, theta


def exp_map(psi, v):
    v_norm = L2norm(v)
    if v_norm.sum() == 0:
        expgam = cos(v_norm) * psi
    else:
        expgam = cos(v_norm) * psi + sin(v_norm) * v / v_norm
    return expgam


def inner_product(psi1, psi2):
    M = psi1.shape[0]
    t = linspace(0,1,M)
    ip = trapz(psi1*psi2, t)
    return ip


def L2norm(psi):
    M = psi.shape[0]
    t = linspace(0,1,M)
    l2norm = sqrt(trapz(psi*psi,t))
    return l2norm


def gam_to_v(gam):
    TT = gam.shape[0]
    time = linspace(0,1,TT)
    binsize = diff(time)
    binsize = binsize.mean()
    mu = ones(TT)
    if gam.ndim == 1:
        psi = sqrt(gradient(gam,binsize))
        vec, theta = inv_exp_map(mu,psi)
    else:
        n = gam.shape[1]

        psi = zeros((TT,n))
        for i in range(0,n):
            psi[:,i] = sqrt(gradient(gam[:, i],binsize))
        
        vec = zeros((TT,n))
        for i in range(0,n):
            out, theta = inv_exp_map(mu,psi[:,i])
            vec[:,i] = out
    
    return vec


def v_to_gam(v):
    TT = v.shape[0]
    time = linspace(0,1,TT)
    mu = ones(TT)
    if v.ndim == 1:
        psi = exp_map(mu,v)
        gam0 = cumtrapz(psi*psi, time, initial=0)
        gam = (gam0 - gam0.min()) / (gam0.max() - gam0.min())
    else:
        n = v.shape[1]

        gam = zeros((TT,n))
        for i in range(0,n):
            psi = exp_map(mu,v[:,i])
            gam0 = cumtrapz(psi*psi, time, initial=0)
            gam[:,i] = (gam0 - gam0.min()) / (gam0.max() - gam0.min())
    
    return(gam)



"""
Warping Invariant PCR Regression using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.curve_functions as cf
import fdasrsf.regression as rg
import fdasrsf.geometry as geo
from scipy.linalg import inv, norm
from scipy.integrate import trapz, cumtrapz

class elastic_curve_pcr_regression:
    """
    This class provides elastic curve pcr regression for functional data using the
    SRVF framework accounting for warping
    
    Usage:  obj = elastic_curve_pcr_regression(beta,y)
    
    :param beta: numpy ndarray of shape (n, M, N) describing N curves in R^M
    :param y: response vector of length N
    :param warp_data: fdacurve object of alignment and shape PCA
    :param alpha: intercept
    :param b: coefficient vector
    :param SSE: sum of squared errors

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021
    """

    def __init__(self, beta, y):
        """
        Construct an instance of the elastic_curve_pcr_regression class
        :param beta: numpy ndarray of shape (n, M, N) describing N curves in R^M
        :param y: response vector
        :param time: vector of size M describing the sample points
        """

        self.beta = beta
        self.y = y

    def calc_model(self, no=5, T=None, rotation=True, parallel=False):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param no: scalar specify number of principal components (default=5)
        :param T: number of resample curve to 
        :param rotation: include rotation (default = T)
        :param parallel: run in parallel (default = F)
        """
        
        self.rotation = rotation
        M = self.beta.shape[1]
        if T is None:
            T = M

        # Align Data
        self.warp_data = fs.fdacurve(self.beta, N=T)
        self.warp_data.karcher_mean(rotation=rotation, parallel=parallel)

        # Calculate PCA
        self.warp_data.shape_pca(no=no)
        
        # OLS using PCA basis
        lam = 0
        R = 0
        N1 = self.warp_data.coef.shape[1]
        Phi = np.ones((N1, no+1))
        Phi[:,1:(no+1)] = self.warp_data.coef.T
        xx = np.dot(Phi.T, Phi)
        inv_xx = inv(xx + lam * R)
        xy = np.dot(Phi.T, self.y)
        b = np.dot(inv_xx, xy)
        alpha = b[0]
        b = b[1:no+1]

        # compute the SSE
        int_X = np.zeros(N1)
        for ii in range(0,N1):
            int_X[ii] = np.sum(self.warp_data.coef[:,ii]*b)
        
        SSE = np.sum((self.y-alpha-int_X)**2)

        self.alpha = alpha
        self.b = b
        self.SSE = SSE

        return
    

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param beta: (n, M,N) matrix of curves
        :param y: truth if available
        """

        T = self.warp_data.beta_mean.shape[1]
        if newdata != None:
            beta = newdata['beta']
            y = newdata['y']
            n = beta.shape[2]
            beta1 = np.zeros(beta.shape)
            q = np.zeros(beta.shape)
            for ii in range(0,n):
                if (beta.shape[1] != T):
                    beta1[:,:,ii] = cf.resamplecurve(beta[:,:,ii],T)
                else:
                    beta1[:,:,ii] = beta[:,:,ii]
                a = -cf.calculatecentroid(beta1[:,:,ii])
                beta1[:,:,ii] += np.tile(a, (T,1)).T
                q[:,:,ii] = cf.curve_to_q(beta1[:,:,ii])[0]
            
            mu = self.warp_data.q_mean

            v = np.zeros(q.shape)
            for ii in range(0,n):
                qn_t, R, gamI = cf.find_rotation_and_seed_unique(mu, q[:,:,ii], 0, self.rotation)
                qn_t = qn_t / np.sqrt(cf.innerprod_q2(qn_t,qn_t))

                q1dotq2 = cf.innerprod_q2(mu,qn_t)

                if (q1dotq2 > 1):
                    q1dotq2 = 1
                
                d = np.arccos(q1dotq2)

                u = qn_t - q1dotq2*mu
                normu = np.sqrt(cf.innerprod_q2(u,u))
                if (normu>1e-4):
                    v[:,:,ii] = u*np.arccos(q1dotq2)/normu
                else:
                    v[:,:,ii] = np.zeros(qn_t.shape)

            
            Utmp = self.warp_data.U.T
            no = self.warp_data.U.shape[1]
            VM = np.mean(self.warp_data.v,2)
            VM = VM.flatten()

            x = np.zeros((no,n))
            for i in range(0,n):
                tmp = v[:,:,i]
                tmpv1 = tmp.flatten()
                x[:,i] = Utmp.dot((tmpv1- VM))

            self.y_pred = np.zeros(n)
            for ii in range(0,n):
                self.y_pred[ii] = self.alpha + np.dot(x[:,ii],self.b)
            
            if y is None:
                self.SSE = np.nan
            else:
                self.SSE = np.sum((y-self.y_pred)**2)
        else:
            n = self.warp_data.coef.shape[1]
            self.y_pred = np.zeros(n)
            for ii in range(0,n):
                self.y_pred[ii] = self.alpha + np.dot(self.warp_data.coef[:,ii],self.b)
            
            self.SSE = np.sum((self.y-self.y_pred)**2)

        return



"""
Warping Invariant PCR Regression using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.fPCA as fpca
import fdasrsf.regression as rg
import fdasrsf.geometry as geo
from scipy.linalg import inv, norm
from scipy.integrate import trapz, cumtrapz
from scipy.optimize import fmin_l_bfgs_b

class elastic_pcr_regression:
    """
    This class provides elastic pcr regression for functional data using the
    SRVF framework accounting for warping
    
    Usage:  obj = elastic_pcr_regression(f,y,time)
    
    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N
    :param warp_data: fdawarp object of alignment
    :param pca: class dependent on fPCA method used object of fPCA
    :param alpha: intercept
    :param b: coefficient vector
    :param SSE: sum of squared errors

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_pcr_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception('Columns of f and time must be equal')

        self.f = f
        self.y = y
        self.time = time

    def calc_model(self, pca_method="combined", no=5, 
                   smooth_data=False, sparam=25, parallel=False, 
                   C=None):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param pca_method: string specifing pca method (options = "combined",
                        "vert", or "horiz", default = "combined")
        :param no: scalar specify number of principal components (default=5)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: run in parallel (default = F)
        :param C: scale balance parameter for combined method (default = None)
        """

        if smooth_data:
            self.f = fs.smooth_data(self.f,sparam)
        
        N1 = self.f.shape[1]

        # Align Data
        self.warp_data = fs.fdawarp(self.f,self.time)
        self.warp_data.srsf_align(parallel=parallel)

        # Calculate PCA
        if pca_method=='combined':
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method=='vert':
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method=='horiz':
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception('Invalid fPCA Method')
        self.pca.calc_fpca(no)
        
        # OLS using PCA basis
        lam = 0
        R = 0
        Phi = np.ones((N1, no+1))
        Phi[:,1:(no+1)] = self.pca.coef
        xx = np.dot(Phi.T, Phi)
        inv_xx = inv(xx + lam * R)
        xy = np.dot(Phi.T, self.y)
        b = np.dot(inv_xx, xy)
        alpha = b[0]
        b = b[1:no+1]

        # compute the SSE
        int_X = np.zeros(N1)
        for ii in range(0,N1):
            int_X[ii] = np.sum(self.pca.coef[ii,:]*b)
        
        SSE = np.sum((self.y-alpha-int_X)**2)

        self.alpha = alpha
        self.b = b
        self.SSE = SSE
        self.pca_method = pca_method

        return
    

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        omethod = self.warp_data.method
        lam = self.warp_data.lam
        M = self.time.shape[0]

        if newdata != None:
            f = newdata['f']
            time = newdata['time']
            y = newdata['y']
            if newdata['smooth']:
                sparam = newdata['sparam']
                f = fs.smooth_data(f,sparam)
            
            q1 = fs.f_to_srsf(f,time)
            n = q1.shape[1]
            self.y_pred = np.zeros(n)
            mq = self.warp_data.mqn
            fn = np.zeros((M,n))
            qn = np.zeros((M,n))
            gam = np.zeros((M,n))
            for ii in range(0,n):
                gam[:,ii] = uf.optimum_reparam(mq,time,q1[:,ii],omethod,lam)
                fn[:,ii] = uf.warp_f_gamma(time,f[:,ii],gam[:,ii])
                qn[:,ii] = uf.f_to_srsf(fn[:,ii],time)
            
            U = self.pca.U
            no = U.shape[1]

            if self.pca.__class__.__name__ == 'fdajpca':
                m_new = np.sign(fn[self.pca.id,:])*np.sqrt(np.abs(fn[self.pca.id,:]))
                qn1 = np.vstack((qn, m_new))
                C = self.pca.C
                TT = self.time.shape[0]
                mu_g = self.pca.mu_g
                mu_psi = self.pca.mu_psi
                vec = np.zeros((M,n))
                psi = np.zeros((TT,n))
                binsize = np.mean(np.diff(self.time))
                for i in range(0,n):
                    psi[:,i] = np.sqrt(np.gradient(gam[:,i],binsize))
                    out, theta = geo.inv_exp_map(mu_psi, psi[:,i])
                    vec[:,i] = out
                
                g = np.vstack((qn1, C*vec))
                a = np.zeros((n,no))
                for i in range(0,n):
                    for j in range(0,no):
                        tmp = (g[:,i]-mu_g)
                        a[i,j] = np.dot(tmp.T, U[:,j])

            elif self.pca.__class__.__name__ == 'fdavpca':
                m_new = np.sign(fn[self.pca.id,:])*np.sqrt(np.abs(fn[self.pca.id,:]))
                qn1 = np.vstack((qn, m_new))
                a = np.zeros((n,no))
                for i in range(0,n):
                    for j in range(0,no):
                        tmp = (qn1[:,i]-self.pca.mqn)
                        a[i,j] = np.dot(tmp.T, U[:,j])

            elif self.pca.__class__.__name__ == 'fdahpca':
                a = np.zeros((n,no))
                mu_psi = self.pca.psi_mu
                vec = np.zeros((M,n))
                TT = self.time.shape[0]
                psi = np.zeros((TT,n))
                binsize = np.mean(np.diff(self.time))
                for i in range(0,n):
                    psi[:,i] = np.sqrt(np.gradient(gam[:,i],binsize))
                    out, theta = geo.inv_exp_map(mu_psi, psi[:,i])
                    vec[:,i] = out
                
                vm = self.pca.vec.mean(axis=1)

                for i in range(0,n):
                    for j in range(0,no):
                        a[i,j] = np.sum(np.dot(vec[:,i]-vm,U[:,j]))
            else: 
                raise Exception('Invalid fPCA Method')

            for ii in range(0,n):
                self.y_pred[ii] = self.alpha + np.dot(a[ii,:],self.b)
            
            if y is None:
                self.SSE = np.nan
            else:
                self.SSE = np.sum((y-self.y_pred)**2)
        else:
            n = self.pca.coef.shape[0]
            self.y_pred = np.zeros(n)
            for ii in range(0,n):
                self.y_pred[ii] = self.alpha + np.dot(self.pca.coef[ii,:],self.b)
            
            self.SSE = np.sum((self.y-self.y_pred)**2)

        return

class elastic_lpcr_regression:
    """
    This class provides elastic logistic pcr regression for functional 
    data using the SRVF framework accounting for warping
    
    Usage:  obj = elastic_lpcr_regression(f,y,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N (-1/1)
    :param warp_data: fdawarp object of alignment
    :param pca: class dependent on fPCA method used object of fPCA
    :param information
    :param alpha: intercept
    :param b: coefficient vector
    :param Loss: logistic loss
    :param PC: probability of classification
    :param ylabels: predicted labels
    
    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_lpcr_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception('Columns of f and time must be equal')

        self.f = f
        self.y = y
        self.time = time

    def calc_model(self, pca_method="combined", no=5, 
                   smooth_data=False, sparam=25, parallel=False):
        """
        This function identifies a logistic regression model with phase-variability
        using elastic pca

        :param pca_method: string specifing pca method (options = "combined",
                        "vert", or "horiz", default = "combined")
        :param no: scalar specify number of principal components (default=5)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: calculate in parallel (default = F)
        :type f: np.ndarray
        :type time: np.ndarray
        """

        if smooth_data:
            self.f = fs.smooth_data(self.f,sparam)
        
        N1 = self.f.shape[1]

        # Align Data
        self.warp_data = fs.fdawarp(self.f,self.time)
        self.warp_data.srsf_align(parallel=parallel)

        # Calculate PCA
        if pca_method=='combined':
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method=='vert':
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method=='horiz':
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception('Invalid fPCA Method')
        self.pca.calc_fpca(no)
        
        # OLS using PCA basis
        lam = 0
        R = 0
        Phi = np.ones((N1, no+1))
        Phi[:,1:(no+1)] = self.pca.coef
        # Find alpha and beta using l_bfgs
        b0 = np.zeros(no+1)
        out = fmin_l_bfgs_b(rg.logit_loss, b0, fprime=rg.logit_gradient,
                            args=(Phi, self.y), pgtol=1e-10, maxiter=200,
                            maxfun=250, factr=1e-30)

        b = out[0]
        alpha = b[0]

        # compute the Loss
        LL = rg.logit_loss(b,Phi,self.y)

        b = b[1:no+1]

        self.alpha = alpha
        self.b = b
        self.LL = LL
        self.pca_method = pca_method

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        omethod = self.warp_data.method
        lam = self.warp_data.lam
        M = self.time.shape[0]

        if newdata != None:
            f = newdata['f']
            time = newdata['time']
            y = newdata['y']
            if newdata['smooth']:
                sparam = newdata['sparam']
                f = fs.smooth_data(f,sparam)
            
            q1 = fs.f_to_srsf(f,time)
            n = q1.shape[1]
            self.y_pred = np.zeros(n)
            mq = self.warp_data.mqn
            fn = np.zeros((M,n))
            qn = np.zeros((M,n))
            gam = np.zeros((M,n))
            for ii in range(0,n):
                gam[:,ii] = uf.optimum_reparam(mq,time,q1[:,ii],omethod)
                fn[:,ii] = uf.warp_f_gamma(time,f[:,ii],gam[:,ii])
                qn[:,ii] = uf.f_to_srsf(fn[:,ii],time)
            
            U = self.pca.U
            no = U.shape[1]

            if self.pca.__class__.__name__ == 'fdajpca':
                m_new = np.sign(fn[self.pca.id,:])*np.sqrt(np.abs(fn[self.pca.id,:]))
                qn1 = np.vstack((qn, m_new))
                C = self.pca.C
                TT = self.time.shape[0]
                mu_g = self.pca.mu_g
                mu_psi = self.pca.mu_psi
                vec = np.zeros((M,n))
                psi = np.zeros((TT,n))
                binsize = np.mean(np.diff(self.time))
                for i in range(0,n):
                    psi[:,i] = np.sqrt(np.gradient(gam[:,i],binsize))
                    out, theta = geo.inv_exp_map(mu_psi, psi[:,i])
                    vec[:,i] = out
                
                g = np.vstack((qn1, C*vec))
                a = np.zeros((n,no))
                for i in range(0,n):
                    for j in range(0,no):
                        tmp = (g[:,i]-mu_g)
                        a[i,j] = np.dot(tmp.T, U[:,j])

            elif self.pca.__class__.__name__ == 'fdavpca':
                m_new = np.sign(fn[self.pca.id,:])*np.sqrt(np.abs(fn[self.pca.id,:]))
                qn1 = np.vstack((qn, m_new))
                a = np.zeros((n,no))
                for i in range(0,n):
                    for j in range(0,no):
                        tmp = (qn1[:,i]-self.pca.mqn)
                        a[i,j] = np.dot(tmp.T, U[:,j])

            elif self.pca.__class__.__name__ == 'fdahpca':
                a = np.zeros((n,no))
                mu_psi = self.pca.psi_mu
                vec = np.zeros((M,n))
                TT = self.time.shape[0]
                psi = np.zeros((TT,n))
                binsize = np.mean(np.diff(self.time))
                for i in range(0,n):
                    psi[:,i] = np.sqrt(np.gradient(gam[:,i],binsize))
                    out, theta = geo.inv_exp_map(mu_psi, psi[:,i])
                    vec[:,i] = out
                
                vm = self.pca.vec.mean(axis=1)

                for i in range(0,n):
                    for j in range(0,no):
                        a[i,j] = np.sum(np.dot(vec[:,i]-vm,U[:,j]))
            else: 
                raise Exception('Invalid fPCA Method')

            for ii in range(0,n):
                self.y_pred[ii] = self.alpha + np.sum(a[ii,:]*self.b)
            
            if y is None:
                self.y_pred = rg.phi(self.y_pred)
                self.y_labels = np.ones(n)
                self.y_labels[self.y_pred < 0.5] = -1
                self.PC = np.nan
            else:
                self.y_pred = rg.phi(self.y_pred)
                self.y_labels = np.ones(n)
                self.y_labels[self.y_pred < 0.5] = -1
                TP = np.sum(y[self.y_labels == 1] == 1)
                FP = np.sum(y[self.y_labels == -1] == 1)
                TN = np.sum(y[self.y_labels == -1] == -1)
                FN = np.sum(y[self.y_labels == 1] == -1)
                self.PC = (TP+TN)/(TP+FP+FN+TN)
        else:
            n = self.pca.coef.shape[0]
            self.y_pred = np.zeros(n)
            for ii in range(0,n):
                self.y_pred[ii] = self.alpha + np.dot(self.pca.coef[ii,:],self.b)
            
            self.y_pred = rg.phi(self.y_pred)
            self.y_labels = np.ones(n)
            self.y_labels[self.y_pred < 0.5] = -1
            TP = np.sum(self.y[self.y_labels == 1] == 1)
            FP = np.sum(self.y[self.y_labels == -1] == 1)
            TN = np.sum(self.y[self.y_labels == -1] == -1)
            FN = np.sum(self.y[self.y_labels == 1] == -1)
            self.PC = (TP+TN)/(TP+FP+FN+TN)

        return
        

class elastic_mlpcr_regression:
    """
    This class provides elastic multinomial logistic pcr regression for functional
    data using the SRVF framework accounting for warping
    
    Usage:  obj = elastic_mlpcr_regression(f,y,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param y: response vector of length N
    :param Y: coded label matrix
    :param warp_data: fdawarp object of alignment
    :param pca: class dependent on fPCA method used object of fPCA
    :param information
    :param alpha: intercept
    :param b: coefficient vector
    :param Loss: logistic loss
    :param PC: probability of classification
    :param ylabels: predicted labels

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_mlpcr_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception('Columns of f and time must be equal')

        self.f = f
        self.y = y
        self.time = time
        N1 = f.shape[1]

        # Code labels
        m = y.max()
        self.n_classes = m
        self.Y = np.zeros((N1, m), dtype=int)
        for ii in range(0, N1):
            self.Y[ii, y[ii]-1] = 1

    def calc_model(self, pca_method="combined", no=5, 
                   smooth_data=False, sparam=25, parallel=False):
        """
        This function identifies a logistic regression model with phase-variability
        using elastic pca

        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: numpy array of N responses
        :param time: vector of size M describing the sample points
        :param pca_method: string specifing pca method (options = "combined",
                        "vert", or "horiz", default = "combined")
        :param no: scalar specify number of principal components (default=5)
        :param smooth_data: smooth data using box filter (default = F)
        :param sparam: number of times to apply box filter (default = 25)
        :param parallel: run model in parallel (default = F)
        :type f: np.ndarray
        :type time: np.ndarray
        """

        if smooth_data:
            self.f = fs.smooth_data(self.f,sparam)
        
        N1 = self.f.shape[1]

        # Align Data
        self.warp_data = fs.fdawarp(self.f,self.time)
        self.warp_data.srsf_align(parallel=parallel)

        # Calculate PCA
        if pca_method=='combined':
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method=='vert':
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method=='horiz':
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception('Invalid fPCA Method')
        self.pca.calc_fpca(no)
        
        # OLS using PCA basis
        lam = 0
        R = 0
        Phi = np.ones((N1, no+1))
        Phi[:,1:(no+1)] = self.pca.coef
        # Find alpha and beta using l_bfgs
        b0 = np.zeros(self.n_classes*(no+1))
        out = fmin_l_bfgs_b(rg.mlogit_loss, b0, fprime=rg.mlogit_gradient,
                                args=(Phi, self.Y), pgtol=1e-10, maxiter=200,
                                maxfun=250, factr=1e-30)

        b = out[0]
        B0 = b.reshape(no+1, self.n_classes)
        alpha = B0[0, :]

        # compute the Loss
        LL = rg.mlogit_loss(b,Phi,self.y)

        b = B0[1:no+1,:]

        self.alpha = alpha
        self.b = b
        self.LL = LL
        self.pca_method = pca_method

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        omethod = self.warp_data.method
        lam = self.warp_data.lam
        m = self.n_classes
        M = self.time.shape[0]

        if newdata != None:
            f = newdata['f']
            time = newdata['time']
            y = newdata['y']
            if newdata['smooth']:
                sparam = newdata['sparam']
                f = fs.smooth_data(f,sparam)
            
            q1 = fs.f_to_srsf(f,time)
            n = q1.shape[1]
            self.y_pred = np.zeros((n,m))
            mq = self.warp_data.mqn
            fn = np.zeros((M,n))
            qn = np.zeros((M,n))
            gam = np.zeros((M,n))
            for ii in range(0,n):
                gam[:,ii] = uf.optimum_reparam(mq,time,q1[:,ii],omethod)
                fn[:,ii] = uf.warp_f_gamma(time,f[:,ii],gam[:,ii])
                qn[:,ii] = uf.f_to_srsf(fn[:,ii],time)
            
            U = self.pca.U
            no = U.shape[1]

            if self.pca.__class__.__name__ == 'fdajpca':
                m_new = np.sign(fn[self.pca.id,:])*np.sqrt(np.abs(fn[self.pca.id,:]))
                qn1 = np.vstack((qn, m_new))
                C = self.pca.C
                TT = self.time.shape[0]
                mu_g = self.pca.mu_g
                mu_psi = self.pca.mu_psi
                vec = np.zeros((M,n))
                psi = np.zeros((TT,n))
                binsize = np.mean(np.diff(self.time))
                for i in range(0,n):
                    psi[:,i] = np.sqrt(np.gradient(gam[:,i],binsize))
                    out, theta = geo.inv_exp_map(mu_psi, psi[:,i])
                    vec[:,i] = out
                
                g = np.vstack((qn1, C*vec))
                a = np.zeros((n,no))
                for i in range(0,n):
                    for j in range(0,no):
                        tmp = (g[:,i]-mu_g)
                        a[i,j] = np.dot(tmp.T, U[:,j])

            elif self.pca.__class__.__name__ == 'fdavpca':
                m_new = np.sign(fn[self.pca.id,:])*np.sqrt(np.abs(fn[self.pca.id,:]))
                qn1 = np.vstack((qn, m_new))
                a = np.zeros((n,no))
                for i in range(0,n):
                    for j in range(0,no):
                        tmp = (qn1[:,i]-self.pca.mqn)
                        a[i,j] = np.dot(tmp.T, U[:,j])

            elif self.pca.__class__.__name__ == 'fdahpca':
                a = np.zeros((n,no))
                mu_psi = self.pca.psi_mu
                vec = np.zeros((M,n))
                TT = self.time.shape[0]
                psi = np.zeros((TT,n))
                binsize = np.mean(np.diff(self.time))
                for i in range(0,n):
                    psi[:,i] = np.sqrt(np.gradient(gam[:,i],binsize))
                    out, theta = geo.inv_exp_map(mu_psi, psi[:,i])
                    vec[:,i] = out
                
                vm = self.pca.vec.mean(axis=1)

                for i in range(0,n):
                    for j in range(0,no):
                        a[i,j] = np.sum(np.dot(vec[:,i]-vm,U[:,j]))
            else: 
                raise Exception('Invalid fPCA Method')

            for ii in range(0,n):
                for jj in range(0,m):
                    self.y_pred[ii,jj] = self.alpha[jj] + np.sum(a[ii,:]*self.b[:,jj])
            
            
            if y is None:
                self.y_pred = rg.phi(self.y_pred.reshape((1,n*m)))
                self.y_pred = self.y_pred.reshape((n,m))
                self.y_labels = np.argmax(self.y_pred,axis=1)
                self.PC = np.nan
            else:
                self.y_pred = rg.phi(self.y_pred.reshape((1,n*m)))
                self.y_pred = self.y_pred.reshape((n,m))
                self.y_labels = np.argmax(self.y_pred,axis=1)
                self.PC = np.zeros(m)
                cls_set = np.arange(0,m)
                for ii in range(0,m):
                    cls_sub = np.setdiff1d(cls_set,ii)
                    TP = np.sum(y[self.y_labels == ii] == ii)
                    FP = np.sum(y[np.in1d(self.y_labels,cls_sub)] == ii)
                    TN = np.sum(y[np.in1d(self.y_labels,cls_sub)] == self.y_labels[np.in1d(self.y_labels,cls_sub)])
                    FN = np.sum(np.in1d(y[self.y_labels==ii],cls_sub))
                    self.PC[ii] = (TP+TN)/(TP+FP+FN+TN)
                
                self.PCo = np.sum(y == self.y_labels)/self.y_labels.shape[0]
        else:
            n = self.pca.coef.shape[0]
            self.y_pred = np.zeros((n,m))
            for ii in range(0,n):
                for jj in range(0,m):
                    self.y_pred[ii,jj] = self.alpha[jj] + np.sum(self.pca.coef[ii,:]*self.b[:,jj])
            
            self.y_pred = rg.phi(self.y_pred.reshape((1,n*m)))
            self.y_pred = self.y_pred.reshape((n,m))
            self.y_labels = np.argmax(self.y_pred,axis=1)
            self.PC = np.zeros(m)
            cls_set = np.arange(0,m)
            for ii in range(0,m):
                cls_sub = np.setdiff1d(cls_set,ii)
                TP = np.sum(self.y[self.y_labels == ii] == ii)
                FP = np.sum(self.y[np.in1d(self.y_labels,cls_sub)] == ii)
                TN = np.sum(self.y[np.in1d(self.y_labels,cls_sub)] == self.y_labels[np.in1d(self.y_labels,cls_sub)])
                FN = np.sum(np.in1d(y[self.y_labels==ii],cls_sub))
                self.PC[ii] = (TP+TN)/(TP+FP+FN+TN)
            
            self.PCo = np.sum(y == self.y_labels)/self.y_labels.shape[0]

        return



"""
Elastic Functional Boxplots

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
import numpy as np
from scipy.integrate import trapz
import fdasrsf.utility_functions as uf
import fdasrsf.geometry as geo
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import collections


class ampbox:
    """
    This class provides amplitude boxplot for functional data using the
    SRVF framework
    
    Usage:  obj = ampbox(warp_data)
    
    :param warp_data: fdawarp class with alignment data
    :type warp_data: fdawarp
    :param Q1: First quartile
    :param Q3: Second quartile
    :param Q1a: First quantile based on alpha
    :param Q3a: Second quantile based on alpha
    :param minn: minimum extreme function
    :param maxx: maximum extreme function
    :param outlier_index: indexes of outlier functions
    :param f_median: median function
    :param q_median: median srvf
    :param plt: surface plot mesh
    
    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the ampbox class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size==0:
            raise Exception('Please align fdawarp class using srsf_align!')

        if fdawarp.type=="mean":
            raise Exception('Please align fdawarp class using srsf_align with method="median"!')

        self.warp_data = fdawarp

    def construct_boxplot(self, alpha=.05, k_a=1):
        """
        This function constructs the amplitude boxplot using the elastic
        square-root slope (srsf) framework.

        :param alpha: quantile value (e.g.,=.05, i.e., 95%)
        :param k_a: scalar for outlier cutoff (e.g.,=1)
        """

        if self.warp_data.rsamps:
            ft = self.warp_data.fs
            f_median = self.warp_data.fmean
            qt = self.warp_data.qs
            q_median = self.warp_data.mqn
            time = self.warp_data.time
        else: 
            ft = self.warp_data.fn
            f_median = self.warp_data.fmean
            qt = self.warp_data.qn
            q_median = self.warp_data.mqn
            time = self.warp_data.time

        N = ft.shape[1]
        lam = 0.5

        # compute amplitude distances
        dy = np.zeros(N)
        for i in range(0,N):
            dy[i] = np.sqrt(trapz((q_median-qt[:,i])**2,time))

        dy_ordering = dy.argsort()
        CR_50 = dy_ordering[0:np.ceil(N/2).astype('int')]
        tmp = dy[CR_50]
        m = tmp.max()

        # identify amplitude quartiles
        angle = np.zeros((CR_50.shape[0],CR_50.shape[0]))
        energy = np.zeros((CR_50.shape[0],CR_50.shape[0]))
        for i in range(0,CR_50.shape[0]-1):
            for j in range(i+1,CR_50.shape[0]):
                q1 = qt[:,CR_50[i]] - q_median
                q3 = qt[:,CR_50[j]] - q_median
                q1 = q1/np.sqrt(trapz(q1**2,time))
                q3 = q3/np.sqrt(trapz(q3**2,time))
                angle[i,j] = trapz(q1*q3,time)
                energy[i,j] = (1-lam) * (dy[CR_50[i]]/m+dy[CR_50[j]]/m) - lam * (angle[i,j] + 1)

        maxloc = energy.argmax()
        maxloc_row,maxloc_col = np.unravel_index(maxloc,energy.shape)

        Q1_index = CR_50[maxloc_row]
        Q3_index = CR_50[maxloc_col]
        Q1_q = qt[:,Q1_index]
        Q3_q = qt[:,Q3_index]
        Q1 = ft[:,Q1_index]
        Q3 = ft[:,Q3_index]

        # identify amplitude quantiles
        dy_ordering = dy.argsort()
        CR_alpha = dy_ordering[0:np.round(N*(1-alpha)).astype('int')]
        tmp = dy[CR_alpha]
        m = tmp.max()
        angle = np.zeros((CR_alpha.shape[0],CR_alpha.shape[0]))
        energy = np.zeros((CR_alpha.shape[0],CR_alpha.shape[0]))
        for i in range(0,CR_alpha.shape[0]-1):
            for j in range(i+1,CR_alpha.shape[0]):
                q1 = qt[:,CR_alpha[i]] - q_median
                q3 = qt[:,CR_alpha[j]] - q_median
                q1 /= np.sqrt(trapz(q1**2,time))
                q3 /= np.sqrt(trapz(q3**2,time))
                angle[i,j] = trapz(q1*q3,time)
                energy[i,j] = (1-lam) * (dy[CR_alpha[i]]/m+dy[CR_alpha[j]]/m) - lam * (angle[i,j] + 1)

        maxloc = energy.argmax()
        maxloc_row,maxloc_col = np.unravel_index(maxloc,energy.shape)

        Q1a_index = CR_alpha[maxloc_row]
        Q3a_index = CR_alpha[maxloc_col]
        Q1a_q = qt[:,Q1a_index]
        Q3a_q = qt[:,Q3a_index]
        Q1a = ft[:,Q1a_index]
        Q3a = ft[:,Q3a_index]

        # compute amplitude whiskers
        IQR = dy[Q1_index] + dy[Q3_index]
        v1 = Q1_q - q_median
        v3 = Q3_q - q_median
        upper_q = Q3_q + k_a * IQR * v3 / np.sqrt(trapz(v3**2,time))
        lower_q = Q1_q + k_a * IQR * v1 / np.sqrt(trapz(v1**2,time))

        upper_dis = np.sqrt(trapz((upper_q - q_median)**2,time))
        lower_dis = np.sqrt(trapz((lower_q-q_median)**2,time))
        whisker_dis = max(upper_dis,lower_dis)

        # identify amplitude outliers
        outlier_index = np.array([])
        for i in range(0,N):
            if dy[dy_ordering[N-i-1]] > whisker_dis:
                outlier_index = np.append(outlier_index,dy[dy_ordering[N+1-i]])
        
        # identify amplitude extremes
        distance_to_upper = np.full(N, np.inf)
        distance_to_lower = np.full(N, np.inf)
        out_50_CR = np.setdiff1d(np.arange(0,N), outlier_index)
        for i in range(0,out_50_CR.shape[0]):
            j = out_50_CR[i]
            distance_to_upper[j] = np.sqrt(trapz((upper_q-qt[:,j])**2,time))
            distance_to_lower[j] = np.sqrt(trapz((lower_q-qt[:,j])**2,time))
        
        max_index = distance_to_upper.argmin()
        min_index = distance_to_lower.argmin()
        min_q = qt[:,min_index]
        max_q = qt[:,max_index]
        minn = ft[:,min_index]
        maxx = ft[:,max_index]

        s = np.linspace(0,1,100)
        Fs2 = np.zeros((time.shape[0],595))
        Fs2[:,0] = (1-s[0]) * minn + s[0] * Q1
        for j in range(1,100):
            Fs2[:,j] = (1-s[j]) * minn + s[j] * Q1a
            Fs2[:,98+j] = (1-s[j]) * Q1a + s[j] * Q1
            Fs2[:,197+j] = (1-s[j]) * Q1 + s[j] * f_median
            Fs2[:,296+j] = (1-s[j]) * f_median + s[j] * Q3
            Fs2[:,395+j] = (1-s[j]) * Q3 + s[j] * Q3a
            Fs2[:,494+j] = (1-s[j]) * Q3a + s[j] * maxx
        
        d1 = np.sqrt(trapz((q_median-Q1_q)**2,time))
        d1a = np.sqrt(trapz((Q1_q-Q1a_q)**2,time))
        dl = np.sqrt(trapz((Q1a_q-min_q)**2,time))
        d3 = np.sqrt(trapz((q_median-Q3_q)**2,time))
        d3a = np.sqrt(trapz((Q3_q-Q3a_q)**2,time))
        du = np.sqrt(trapz((Q3a_q-max_q)**2,time))
        part1=np.linspace(-d1-d1a-dl,-d1-d1a,100)
        part2=np.linspace(-d1-d1a,-d1,100)
        part3=np.linspace(-d1,0,100)
        part4=np.linspace(0,d3,100)
        part5=np.linspace(d3,d3+d3a,100)
        part6=np.linspace(d3+d3a,d3+d3a+du,100)
        allparts = np.hstack((part1,part2[1:100],part3[1:100],part4[1:100],part5[1:100],part6[1:100]))
        U, V = np.meshgrid(time, allparts)
        U = np.transpose(U)
        V = np.transpose(V)

        self.Q1 = Q1
        self.Q3 = Q3
        self.Q1a = Q1a
        self.Q3a = Q3a
        self.minn = minn
        self.maxx = maxx
        self.outlier_index = outlier_index
        self.f_median = f_median
        self.q_median = q_median
        
        plt = collections.namedtuple('plt', ['U', 'V', 'Fs2', 'allparts',
                                                        'd1', 'd1a', 'dl',
                                                        'd3', 'd3a', 'du',
                                                        'Q1q','Q3q'])

        self.plt = plt(U,V,Fs2,allparts,d1,d1a,dl,d3,d3a,du,Q1a_q,Q3a_q)                                           
        return

    def plot(self):
        """
        plot box plot and surface plot

        Usage: obj.plot()
        """
        M = self.warp_data.time.shape[0]
        fig1, ax1 = plt.subplots()
        ax1.plot(self.warp_data.time,self.f_median,'k')
        ax1.plot(self.warp_data.time,self.Q1,'b')
        ax1.plot(self.warp_data.time,self.Q3,'b')
        ax1.plot(self.warp_data.time,self.Q1a,'g')
        ax1.plot(self.warp_data.time,self.Q3a,'g')
        ax1.plot(self.warp_data.time,self.minn,'r')
        ax1.plot(self.warp_data.time,self.maxx,'r')
        
        fig2 = plt.figure() 
        ax = fig2.gca(projection='3d')
        ax.plot_surface(self.plt.U,self.plt.V,self.plt.Fs2,cmap=cm.viridis,rcount=200,ccount=200)
        ax.plot(self.warp_data.time, np.zeros(M), self.f_median,'k')
        ax.plot(self.warp_data.time, np.repeat(-self.plt.d1,M), self.Q1,'b')
        ax.plot(self.warp_data.time, np.repeat(-self.plt.d1-self.plt.d1a,M),self.Q1a,'g')
        ax.plot(self.warp_data.time, np.repeat(-self.plt.d1-self.plt.d1a-self.plt.dl,M),self.minn,'r')
        ax.plot(self.warp_data.time, np.repeat(self.plt.d3,M),self.Q3,'b')
        ax.plot(self.warp_data.time, np.repeat(self.plt.d3+self.plt.d3a,M),self.Q3a,'g')
        ax.plot(self.warp_data.time, np.repeat(self.plt.d3+self.plt.d3a+self.plt.du,M),self.maxx,'r')

        plt.show()


class phbox:
    """
    This class provides phase boxplot for functional data using the
    SRVF framework
    
    Usage:  obj = phbox(warp_data)
    
    :param warp_data: fdawarp class with alignment data
    :type warp_data: fdawarp
    :param Q1: First quartile
    :param Q3: Second quartile
    :param Q1a: First quantile based on alpha
    :param Q3a: Second quantile based on alpha
    :param minn: minimum extreme function
    :param maxx: maximum extreme function
    :param outlier_index: indexes of outlier functions
    :param median_x: median warping function
    :param psi_median: median srvf of warping function
    :param plt: surface plot mesh
    
    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the phbox class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size==0:
            raise Exception('Please align fdawarp class using srsf_align!')

        if fdawarp.type=="mean":
            raise Exception('Please align fdawarp class using srsf_align with method="median"!')

        self.warp_data = fdawarp

    def construct_boxplot(self, alpha=.05, k_a=1):
        """
        This function constructs phase boxplot for functional data using the elastic
        square-root slope (srsf) framework.

        :param alpha: quantile value (e.g.,=.05, i.e., 95%)
        :param k_a: scalar for outlier cutoff (e.g.,=1)
        """

        if self.warp_data.rsamps:
            gam = self.warp_data.gams
        else:
            gam = self.warp_data.gam
        M,N = gam.shape
        t = np.linspace(0,1,M)
        time = t
        lam = 0.5

        # compute phase median
        median_x, psi_median, psi, vec = uf.SqrtMedian(gam)

        # compute phase distances
        dx = np.zeros(N)
        v = np.zeros((M,N))
        for k in range(0, N):
            v[:,k], d = geo.inv_exp_map(psi_median,psi[:,k])
            dx[k] = np.sqrt(trapz(v[:,k]**2,t))

        dx_ordering = dx.argsort()
        CR_50 = dx_ordering[0:np.ceil(N/2).astype('int')]
        tmp = dx[CR_50]
        m = tmp.max()

        # identify phase quartiles
        angle = np.zeros((CR_50.shape[0],CR_50.shape[0]))
        energy = np.zeros((CR_50.shape[0],CR_50.shape[0]))
        for i in range(0,CR_50.shape[0]-1):
            for j in range(i+1,CR_50.shape[0]):
                q1 = v[:,CR_50[i]]
                q3 = v[:,CR_50[j]]
                q1 /= np.sqrt(trapz(q1**2,time))
                q3 /= np.sqrt(trapz(q3**2,time))
                angle[i,j] = trapz(q1*q3,time)
                energy[i,j] = (1-lam) * (dx[CR_50[i]]/m+dx[CR_50[j]]/m) - lam * (angle[i,j] + 1)

        maxloc = energy.argmax()
        maxloc_row, maxloc_col = np.unravel_index(maxloc,energy.shape)

        Q1_index = CR_50[maxloc_row]
        Q3_index = CR_50[maxloc_col]
        Q1 = gam[:,Q1_index]
        Q3 = gam[:,Q3_index]
        Q1_psi = np.sqrt(np.gradient(Q1,1/(M-1)))
        Q3_psi = np.sqrt(np.gradient(Q3,1/(M-1)))

        # identify phase quantiles
        dx_ordering = dx.argsort()
        CR_alpha = dx_ordering[0:np.round(N*(1-alpha)).astype('int')]
        tmp = dx[CR_alpha]
        m = tmp.max()
        angle = np.zeros((CR_alpha.shape[0],CR_alpha.shape[0]))
        energy = np.zeros((CR_alpha.shape[0],CR_alpha.shape[0]))
        for i in range(0,CR_alpha.shape[0]-1):
            for j in range(i+1,CR_alpha.shape[0]):
                q1 = v[:,CR_alpha[i]]
                q3 = v[:,CR_alpha[j]]
                q1 /= np.sqrt(trapz(q1**2,time))
                q3 /= np.sqrt(trapz(q3**2,time))
                angle[i,j] = trapz(q1*q3,time)
                energy[i,j] = (1-lam) * (dx[CR_alpha[i]]/m+dx[CR_alpha[j]]/m) - lam * (angle[i,j] + 1)

        maxloc = energy.argmax()
        maxloc_row, maxloc_col = np.unravel_index(maxloc,energy.shape)

        Q1a_index = CR_alpha[maxloc_row]
        Q3a_index = CR_alpha[maxloc_col]
        Q1a = gam[:,Q1a_index]
        Q3a = gam[:,Q3a_index]
        Q1a_psi = np.sqrt(np.gradient(Q1a,1/(M-1)))
        Q3a_psi = np.sqrt(np.gradient(Q3a,1/(M-1)))

        # check quartile and quantile going in same direction
        tst = trapz(v[:,Q1a_index]*v[:,Q1_index])
        if tst < 0:
            Q1a = gam[:,Q3a_index]
            Q3a = gam[:,Q1a_index]

        # compute phase whiskers
        IQR = dx[Q1_index] + dx[Q3_index]
        v1 = v[:,Q3a_index]
        v3 = v[:,Q3a_index]
        upper_v = v3 + k_a * IQR * v3 / np.sqrt(trapz(v3**2,time))
        lower_v = v1 + k_a * IQR * v1 / np.sqrt(trapz(v1**2,time))

        upper_dis = np.sqrt(trapz(v3**2,time))
        lower_dis = np.sqrt(trapz(v1**2,time))
        whisker_dis = max(upper_dis,lower_dis)

        # identify phase outliers
        outlier_index = np.array([])
        for i in range(0,N):
            if dx[dx_ordering[N-1-i]] > whisker_dis:
                outlier_index = np.append(outlier_index,dx_ordering[N+1-i])
        
        # identify phase extremes
        distance_to_upper = np.full(N, np.inf)
        distance_to_lower = np.full(N, np.inf)
        out_50_CR = np.setdiff1d(np.arange(0,N), outlier_index)
        for i in range(0,out_50_CR.shape[0]):
            j = out_50_CR[i]
            distance_to_upper[j] = np.sqrt(trapz((upper_v-v[:,j])**2,time))
            distance_to_lower[j] = np.sqrt(trapz((lower_v-v[:,j])**2,time))
        
        max_index = distance_to_upper.argmin()
        min_index = distance_to_lower.argmin()
        minn = gam[:,min_index]
        maxx = gam[:,max_index]
        min_psi = psi[:,min_index]
        max_psi = psi[:,max_index]

        s = np.linspace(0,1,100)
        Fs2 = np.zeros((time.shape[0],595))
        Fs2[:,0] = (1-s[0]) * (minn-t) + s[0] * (Q1-t)
        for j in range(1,100):
            Fs2[:,j] = (1-s[j]) * (minn-t) + s[j] * (Q1a-t)
            Fs2[:,98+j] = (1-s[j]) * (Q1a-t) + s[j] * (Q1-t)
            Fs2[:,197+j] = (1-s[j]) * (Q1-t) + s[j] * (median_x-t)
            Fs2[:,296+j] = (1-s[j]) * (median_x-t) + s[j] * (Q3-t)
            Fs2[:,395+j] = (1-s[j]) * (Q3-t) + s[j] * (Q3a-t)
            Fs2[:,494+j] = (1-s[j]) * (Q3a-t) + s[j] * (maxx-t)
        
        d1 = np.sqrt(trapz(psi_median*Q1_psi,time))
        d1a = np.sqrt(trapz(Q1_psi*Q1a_psi,time))
        dl = np.sqrt(trapz(Q1a_psi*min_psi,time))
        d3 = np.sqrt(trapz((psi_median*Q3_psi),time))
        d3a = np.sqrt(trapz((Q3_psi*Q3a_psi),time))
        du = np.sqrt(trapz((Q3a_psi*max_psi),time))
        part1=np.linspace(-d1-d1a-dl,-d1-d1a,100)
        part2=np.linspace(-d1-d1a,-d1,100)
        part3=np.linspace(-d1,0,100)
        part4=np.linspace(0,d3,100)
        part5=np.linspace(d3,d3+d3a,100)
        part6=np.linspace(d3+d3a,d3+d3a+du,100)
        allparts = np.hstack((part1,part2[1:100],part3[1:100],part4[1:100],part5[1:100],part6[1:100]))
        U, V = np.meshgrid(time, allparts)
        U = np.transpose(U)
        V = np.transpose(V)

        self.Q1 = Q1
        self.Q3 = Q3
        self.Q1a = Q1a
        self.Q3a = Q3a
        self.minn = minn
        self.maxx = maxx
        self.outlier_index = outlier_index
        self.median_x = median_x
        self.psi_media = psi_median
        
        plt = collections.namedtuple('plt', ['U', 'V', 'Fs2', 'allparts',
                                                        'd1', 'd1a', 'dl',
                                                        'd3', 'd3a', 'du',
                                                        'Q1_psi','Q3_psi'])

        self.plt = plt(U,V,Fs2,allparts,d1,d1a,dl,d3,d3a,du,Q1a_psi,Q3a_psi)                                           
        return
    
    def plot(self):
        """
        plot box plot and surface plot

        Usage: obj.plot()
        """

        M = self.warp_data.time.shape[0]
        time = np.linspace(0,1,M)
        fig1, ax1 = plt.subplots()
        ax1.plot(time,self.median_x,'k')
        ax1.plot(time,self.Q1,'b')
        ax1.plot(time,self.Q3,'b')
        ax1.plot(time,self.Q1a,'g')
        ax1.plot(time,self.Q3a,'g')
        ax1.plot(time,self.minn,'r')
        ax1.plot(time,self.maxx,'r')
        ax1.set_aspect('equal')
        
        fig2 = plt.figure() 
        ax = fig2.gca(projection='3d')
        ax.plot_surface(self.plt.U,self.plt.V,self.plt.Fs2,cmap=cm.viridis,rcount=200,ccount=200)
        ax.plot(time, np.zeros(M), self.median_x-time,'k')
        ax.plot(time, np.repeat(-self.plt.d1,M), self.Q1 - time,'b')
        ax.plot(time, np.repeat(-self.plt.d1-self.plt.d1a,M),self.Q1a - time,'g')
        ax.plot(time, np.repeat(-self.plt.d1-self.plt.d1a-self.plt.dl,M),self.minn - time,'r')
        ax.plot(time, np.repeat(self.plt.d3,M),self.Q3 - time,'b')
        ax.plot(time, np.repeat(self.plt.d3+self.plt.d3a,M),self.Q3a - time,'g')
        ax.plot(time, np.repeat(self.plt.d3+self.plt.d3a+self.plt.du,M),self.maxx - time,'r')

        plt.show()

        return



"""
statistic calculation for SRVF (curves) open and closed using Karcher
Mean and Variance

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
from numpy import zeros, sqrt, fabs, cos, sin, tile, vstack, empty, cov, inf, mean, arange, prod
from numpy import append, arccos, cumsum, sum
from numpy.linalg import svd
from numpy.random import randn
import fdasrsf.curve_functions as cf
import fdasrsf.utility_functions as uf
import fdasrsf.plot_style as plot
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
import collections

class fdacurve:
    """
    This class provides alignment methods for open and closed curves using the SRVF framework

    Usage: obj = fdacurve(beta, mode, N, scale)
    :param beta: numpy ndarray of shape (n, M, N) describing N curves
    in R^M
    :param mode: Open ('O') or closed curve ('C') (default 'O')
    :param N: resample curve to N points
    :param scale: scale curve to length 1 (true/false)
    :param q:           (n,T,K) matrix defining n dimensional srvf on T samples with K srvfs
    :param betan:       aligned curves
    :param qn:          aligned srvfs
    :param basis:       calculated basis
    :param beta_mean:   karcher mean curve
    :param q_mean:      karcher mean srvf
    :param gams:        warping functions
    :param v:           shooting vectors
    :param C:           karcher covariance
    :param s:           pca singular values
    :param U:           pca singular vectors
    :param coef:        pca coefficients
    :param pca          principal directions
    :param qun:         cost function
    :param samples:     random samples
    :param gamr:        random warping functions
    :param cent:        center
    :param scale:       scale
    :param len:         length of curve
    :param len_q:       length of srvf
    :param mean_scale   mean length
    :param mean_scale_q mean length srvf
    :param E:           energy

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  26-Aug-2020
    """

    def __init__(self, beta, mode='O', N=200, scale=False):
        """
        fdacurve Construct an instance of this class
        :param beta: (n,T,K) matrix defining n dimensional curve on T samples with K curves
        :param mode:  Open ('O') or closed curve ('C') (default 'O')
        :param N: resample curve to N points
        :param scale: include scale (true/false)
        """
        self.mode = mode
        self.scale = scale

        K = beta.shape[2]
        n = beta.shape[0]
        q = zeros((n,N,K))
        beta1 = zeros((n,N,K))
        cent1 = zeros((n,K))
        len1 = zeros(K)
        lenq1 = zeros(K)
        for ii in range(0,K):
            if (beta.shape[1] != N):
                beta1[:,:,ii] = cf.resamplecurve(beta[:,:,ii],N,mode=mode)
            else:
                beta1[:,:,ii] = beta[:,:,ii]
            a = -cf.calculatecentroid(beta1[:,:,ii])
            beta1[:,:,ii] += tile(a, (N,1)).T
            q[:,:,ii], len1[ii], lenq1[ii] = cf.curve_to_q(beta1[:,:,ii], mode)
            cent1[:,ii] = -a

        self.q = q
        self.beta = beta1
        self.cent = cent1
        self.len = len1
        self.len_q = lenq1


    def karcher_mean(self, rotation=True, parallel=False, cores=-1, method="DP"):
        """
        This calculates the mean of a set of curves
        :param rotation: compute optimal rotation (default = T)
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"
        """
        n, T, N = self.beta.shape

        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        # Initialize mu as one of the shapes
        mu = self.q[:, :, 0]
        betamean = self.beta[:,:,0]
        itr = 0

        gamma = zeros((T,N))
        maxit = 20

        sumd = zeros(maxit+1)
        v = zeros((n, T, N))
        sumv = zeros((n,T))
        normvbar = zeros(maxit+1)

        delta = 0.5
        tolv = 1e-4
        told = 5*1e-3

        print("Computing Karcher Mean of %d curves in SRVF space.." % N)
        while itr < maxit:
            print("updating step: %d" % (itr+1))

            if iter == maxit:
                print("maximal number of iterations reached")

            mu = mu / sqrt(cf.innerprod_q2(mu, mu))
            if mode == 1:
                self.basis = cf.find_basis_normal(mu)
            else:
                self.basis = []

            sumv = zeros((n, T))
            sumd[0] = inf
            sumd[itr+1] = 0
            out = Parallel(n_jobs=cores)(delayed(karcher_calc)(mu, self.q[:, :, n], self.basis, 
                                                               mode, rotation, method) for n in range(N))
            v = zeros((n, T, N))
            gamma = zeros((T,N))
            for i in range(0, N):
                v[:, :, i] = out[i][0]
                gamma[:, i] = out[i][1]
                sumv += v[:, :, i]
                sumd[itr+1] = sumd[itr+1] + out[i][2]**2

            sumv = v.sum(axis=2)

            # Compute average direction of tangent vectors v_i
            vbar = sumv/float(N)

            normvbar[itr] = sqrt(cf.innerprod_q2(vbar, vbar))
            normv = normvbar[itr]

            if ((sumd[itr]-sumd[itr+1]) < 0):
                break
            elif (normv > tolv and fabs(sumd[itr+1]-sumd[itr]) > told):
                # Update mu in direction of vbar
                mu = cos(delta*normvbar[itr])*mu + sin(delta*normvbar[itr]) * vbar/normvbar[itr]

                if mode == 1:
                    mu = cf.project_curve(mu)

                x = cf.q_to_curve(mu)
                a = -1*cf.calculatecentroid(x)
                betamean = x + tile(a, [T, 1]).T
            else:
                break

            itr += 1

        # compute average length
        if self.scale:
            self.mean_scale = (prod(self.len))**(1./self.len.shape[0])
            self.mean_scale_q = (prod(self.len_q))**(1./self.len.shape[0])
            betamean = self.mean_scale * betamean
        
        self.q_mean = mu
        self.beta_mean = betamean
        self.v = v
        self.qun = sumd[0:(itr+1)]
        self.E = normvbar[0:(itr+1)]

        return


    def srvf_align(self, rotation=True, parallel=False, cores=-1, method="DP"):
        """
        This aligns a set of curves to the mean and computes mean if not computed
        :param rotation: compute optimal rotation (default = T)
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        :param method: method to apply optimization (default="DP") options are "DP" or "RBFGS"
        """
        n, T, N = self.beta.shape

        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        # find mean
        if not hasattr(self, 'beta_mean'):
            self.karcher_mean()

        self.qn = zeros((n, T, N))
        self.betan = zeros((n, T, N))
        self.gams = zeros((T,N))
        centroid2 = cf.calculatecentroid(self.beta_mean)
        self.beta_mean = self.beta_mean - tile(centroid2, [T, 1]).T

        # align to mean
        out = Parallel(n_jobs=-1)(delayed(cf.find_rotation_and_seed_unique)(self.q_mean,
                                          self.q[:, :, n], mode, rotation, method) for n in range(N))
        for ii in range(0, N):
            self.gams[:,ii] = out[ii][2]
            self.qn[:, :, ii] = out[ii][0]
            btmp = out[ii][1].dot(self.beta[:,:,ii])
            self.betan[:, :, ii] = cf.group_action_by_gamma_coord(btmp,out[ii][2])

        return


    def karcher_cov(self):
        """
        This calculates the mean of a set of curves

        """
        if not hasattr(self, 'beta_mean'):
            self.karcher_mean()

        M,N,K = self.v.shape

        N1 = M*N
        if (self.scale):
            N1 += 1

        tmpv = zeros((N1,K))
        for i in range(0,K):
            tmp = self.v[:,:,i]
            tmpv1 = tmp.flatten()
            if (self.scale):
                tmpv1 = append(tmpv1, self.len[i])

            tmpv[:,i] = tmpv1

        if (self.scale):
            VM = mean(self.v, 2)
            VM = VM.flatten()
            VM = append(VM, self.mean_scale)
            tmpv = tmpv - tile(VM, [K, 1]).T
            self.C = cov(tmpv)
        else:
            self.C = cov(tmpv)

        return


    def shape_pca(self, no=10):
        """
        Computes principal direction of variation specified by no. N is
        Number of shapes away from mean. Creates 2*N+1 shape sequence

        :param no: number of direction (default 3)
        """
        if not hasattr(self, 'C'):
            self.karcher_cov()

        U1, s, V = svd(self.C)
        self.U = U1[:,0:no]
        self.s = s[0:no]

        # express shapes as coefficients
        K = self.beta.shape[2]
        VM = mean(self.v,2)
        VM = VM.flatten()
        if (self.scale):
            VM = append(VM, self.mean_scale)

        x = zeros((no,K))
        for ii in range(0,K):
            tmpv = self.v[:,:,ii]
            tmpv1 = tmpv.flatten()
            if (self.scale):
                tmpv1 = append(tmpv1, self.len[ii])
            Utmp = self.U.T
            x[:,ii] = Utmp.dot((tmpv1- VM))

        self.coef = x

        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        n1, T, N1 = self.beta.shape
        p = zeros((n1,T,no,10))
        for j in range(0,no):
            for i in range(1,11):
                tmp = VM + 0.5*(i-5)*sqrt(self.s[j])*self.U[:,j]
                m,n = self.q_mean.shape
                if (self.scale):
                    tmp_scale = tmp[-1]
                    tmp = tmp[0:-1]
                else:
                    tmp_scale = 1
                v1 = tmp.reshape(m,n)
                q2n = cf.elastic_shooting(self.q_mean, v1, mode)

                p[:,:,j,i-1] = cf.q_to_curve(q2n, tmp_scale)            

        self.pca = p

        return


    def sample_shapes(self, no=3, numSamp=10):
        """
        Computes sample shapes from mean and covariance

        :param no: number of direction (default 3)
        :param numSamp: number of samples (default 10)
        """
        n, T = self.q_mean.shape
        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        U, s, V = svd(self.C)

        if mode == 0:
            N = 2
        else:
            N = 10

        epsilon = 1./(N-1)

        samples = empty(numSamp, dtype=object)
        for i in range(0, numSamp):
            v = zeros((2, T))
            for m in range(0, no):
                v = v + randn()*sqrt(s[m])*vstack((U[0:T, m], U[T:2*T, m]))

            q1 = self.q_mean
            for j in range(0, N-1):
                normv = sqrt(cf.innerprod_q2(v, v))

                if normv < 1e-4:
                    q2 = self.q_mean
                else:
                    q2 = cos(epsilon*normv)*q1+sin(epsilon*normv)*v/normv
                    if mode == 1:
                        q2 = cf.project_curve(q2)

                # Parallel translate tangent vector
                basis2 = cf.find_basis_normal(q2)
                v = cf.parallel_translate(v, q1, q2, basis2, mode)

                q1 = q2

            samples[i] = cf.q_to_curve(q2)

        self.samples = samples
        return


    def plot(self):
        """
        plot curve mean results
        """
        fig, ax = plt.subplots()
        n,T,K = self.beta.shape
        for ii in range(0,K):
            ax.plot(self.beta[0,:,ii],self.beta[1,:,ii])
        plt.title('Curves')
        ax.set_aspect('equal')
        plt.axis('off')
        plt.gca().invert_yaxis()

        if hasattr(self,'gams'):
            M = self.gams.shape[0]
            fig, ax = plot.f_plot(arange(0, M) / float(M - 1), self.gams,
                                    title="Warping Functions")

        if hasattr(self,'beta_mean'):
            fig, ax = plt.subplots()
            ax.plot(self.beta_mean[0,:],self.beta_mean[1,:])
            plt.title('Karcher Mean')
            ax.set_aspect('equal')
            plt.axis('off')
            plt.gca().invert_yaxis()

        plt.show()


    def plot_pca(self):

        if not hasattr(self,'s'):
            raise NameError('Calculate PCA')

        fig, ax = plt.subplots()
        ax.plot(cumsum(self.s)/sum(self.s)*100)
        plt.title('Variability Explained')
        plt.xlabel('PC')

        # plot principal modes of variability
        VM = mean(self.v,2)
        VM = VM.flatten()
        if (self.scale):
            VM = append(VM, self.mean_scale)

        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        for j in range(0,4):
            fig, ax = plt.subplots()
            for i in range(1,11):
                tmp = VM + 0.5*(i-5)*sqrt(self.s[j])*self.U[:,j]
                m,n = self.q_mean.shape
                if (self.scale):
                    tmp_scale = tmp[-1]
                    tmp = tmp[0:-1]
                else:
                    tmp_scale = 1
                v1 = tmp.reshape(m,n)
                q2n = cf.elastic_shooting(self.q_mean,v1,mode)

                p = cf.q_to_curve(q2n, tmp_scale)

                mv = 0.2
                if (self.scale):
                    mv *= self.mean_scale
                
                if i == 5:
                    ax.plot(mv*i + p[0,:], p[1,:], 'k', linewidth=2)
                else:
                    ax.plot(mv*i + p[0,:], p[1,:], linewidth=2)

            ax.set_aspect('equal')
            plt.axis('off')
            plt.title('PD %d' % (j+1))

        plt.show()


def karcher_calc(mu, q, basis, closed, rotation, method):
    # Compute shooting vector from mu to q_i
    qn_t, R, gamI = cf.find_rotation_and_seed_unique(mu, q, closed, rotation, method)
    qn_t = qn_t / sqrt(cf.innerprod_q2(qn_t,qn_t))

    q1dotq2 = cf.innerprod_q2(mu,qn_t)

    if (q1dotq2 > 1):
        q1dotq2 = 1
    
    d = arccos(q1dotq2)

    u = qn_t - q1dotq2*mu
    normu = sqrt(cf.innerprod_q2(u,u))
    if (normu>1e-4):
        w = u*arccos(q1dotq2)/normu
    else:
        w = zeros(qn_t.shape)

    # Project to tangent space of manifold to obtain v_i
    if closed == 0:
        v = w
    else:
        v = cf.project_tangent(w, q, basis)

    return(v, gamI, d)
